#include <cstdlib>
#include <filesystem>
#include "utils/command_parser.hpp"
#include "projects/project_list.hpp"
#include "utils/local.hpp"
#include "lua/loader.hpp"

#include <tmuxub/tmuxub.hpp>

namespace fs = std::filesystem;

void OpenFileInEditor(const std::filesystem::path& filename) {
  const char* editor = getenv("EDITOR");

  if (!editor) {
    editor = "vim";
  }

  auto str = filename.string();

  std::vector<const char*> argv = {editor, str.data(), nullptr};
  execvp(editor, const_cast<char* const*>(argv.data()));
}

int main(int argc, char* argv[]) {
  ProjectList projects;
  projects.LoadFromYaml(GetLocalPath() / "projects.yml");

  LuaLoader lua;
  lua.AddLuaPath((GetLocalPath() / "globals/") / "?.lua");
  lua.AddCPath((GetLocalPath() / "globals/") / "?.so");

  Parser parser;

  try {
    const auto& project = projects.GetProject();
    lua.AddLuaPath(project.GetPathToScripts() / "?.lua");

    std::vector<fs::path> scripts;
    for (auto entry : fs::directory_iterator(project.GetPathToScripts())) {
      auto filename = entry.path().filename().string();
      if (filename.ends_with(".lua")) {
        scripts.push_back(filename.substr(0, filename.size() - 4));
      }
    }

    for (auto script_name : scripts) {
      lua.LoadToGlobal(script_name);

      parser.AddRule(MakePattern({".", "config", script_name.string().data()}),
                     [&projects](auto& args) {
                       const auto& project = projects.GetProject();
                       OpenFileInEditor(project.GetPathToScripts() / (args[3] + ".lua"));
                     });

      if (script_name == "main") {
        auto all_scripts = lua.GetListScripts("main");

        for (auto& script : all_scripts) {
          parser.AddRule(MakePattern({script.data()}), [script, &lua, &project](auto& args) {
            fs::current_path(project.GetPath());
            lua.Call("main", script);
          });
        }
      } else {
        auto all_scripts = lua.GetListScripts(script_name);

        for (auto& script : all_scripts) {
          parser.AddRule(MakePattern({script_name.string().data(), script.data()}),
                         [script, &lua, &project, script_name](auto& args) {
                           std::filesystem::current_path(project.GetPath());
                           lua.Call(script_name, script);
                         });
        }
      }
    }
  } catch (...) {
  }

  for (auto& project : projects.GetProjects()) {
    parser.AddRule(MakePattern({".", "open", project.GetName().data()}), [&project](auto& args) {
      tmuxub::Tmux t(GetLocalPath() / "tmuxsocket");

      if (t.HasSession(project.GetName())) {
        t.ConnectToSession(project.GetName());
        t.Exec();
      } else {
        fs::current_path(project.GetPath());
        t.CreateSession(project.GetName());
        t.Exec();
      }

      std::cout << "try open: " << project.GetName() << std::endl;
    });
  }

  parser.AddRule(MakePattern({"help"}), [](auto& args) { std::cout << "It's help" << std::endl; });

  parser.AddRule(MakePattern({".", "new"}),
                 [&projects](auto& args) { projects.AddProject(CreateProject()); });

  parser.AddRule(MakePattern({".", "new", PatternType::AnyWord}),
                 [&projects](auto& args) { projects.AddProject(CreateProject(args[3])); });

  parser.AddRule(MakePattern({"."}), [&projects](auto& args) {
    const auto& project = projects.GetProject();
    OpenFileInEditor((project.GetPathToScripts() / "main.lua").string());
  });

  parser.AddRule(MakePattern({".", "global", PatternType::AnyWord}), [&projects](auto& args) {
    OpenFileInEditor(GetLocalPath() / "globals" / (args[3] + ".lua"));
  });

  parser.AddRule(MakePattern({".", "config"}), [&projects](auto& args) {
    const auto& project = projects.GetProject();
    OpenFileInEditor((project.GetPathToScripts() / "main.lua").string());
  });

  parser.AddRule(MakePattern({".", "config", PatternType::AnyWord}), [&projects](auto& args) {
    const auto& project = projects.GetProject();
    OpenFileInEditor(project.GetPathToScripts() / (args[3] + ".lua"));
  });

  parser.AddRule(MakePattern({".", "projects"}),
                 [](auto& args) { OpenFileInEditor(GetLocalPath() / "projects.yml"); });

  parser.AddRule(MakePattern({PatternType::AnyWords}),
                 [](auto& args) { std::cout << "Unknown command" << std::endl; });

  if (getenv("CL_SUGGESTION") != nullptr) {
    auto suggestions = parser.GetPossibleAdditions(argc, argv);
    for (auto& suggestion : suggestions) {
      std::cout << suggestion << " ";
    }
  } else {
    parser.Parse(argc, argv);

    projects.SaveToYaml(GetLocalPath() / "projects.yml");
  }
}
