#include "utils/command_parser.hpp"
#include <deque>

void Parser::AddRuleImpl(const std::vector<Pattern>& pattern, const Callback& callback,
                         bool find_in_suggestion) {
  rules_.emplace_back(pattern, callback, find_in_suggestion);
}

void Parser::Parse(int argc, char* argv[]) const {
  std::vector<std::string> args(argv, argv + argc);

  for (const auto& [pattern, callback, _] : rules_) {
    if (Match(pattern, args)) {
      if (callback.index() == 0) {
        std::get<0>(callback)(args);
        return;
      } else {
        if (std::get<1>(callback)(args)) {
          return;
        }
      }
    }
  }
}

std::vector<std::string> Parser::GetPossibleAdditions(int argc, char* argv[]) const {
  if (argc != 2) {
    throw std::logic_error("argc should be equal 2");
  }

  size_t last = 0;
  std::string s = argv[1];

  std::vector<std::string> input;

  for (size_t i = 0; i < s.length(); ++i) {
    if (s[i] == ' ') {
      input.push_back(s.substr(last, i - last));
      last = i + 1;
    } else if (i + 1 == s.length()) {
      input.push_back(s.substr(last, i - last + 1));
    }
  }

  std::erase_if(input, [](auto x) { return x.empty(); });

  if (s.back() == ' ') {
    input.emplace_back();
  }

  return GetPossibleAdditions(input);
};

std::vector<std::string> Parser::GetPossibleAdditions(const std::vector<std::string>& input) const {
  std::vector<std::string> suggestions;

  for (const auto& [pattern, callback, check] : rules_) {
    if (!check) {
      continue;
    }
    size_t arg_index = 1;
    size_t pattern_index = 0;
    bool continue_outer_loop = false;

    while (arg_index <= input.size() && pattern_index < pattern.size()) {
      switch (pattern[pattern_index].type) {
        case PatternType::FixedWord:
          if (arg_index == input.size()) {
            suggestions.push_back(pattern[pattern_index].fixed_word);
            continue_outer_loop = true;
          } else if (arg_index + 1 == input.size() &&
                     pattern[pattern_index].fixed_word.starts_with(input[arg_index])) {
            suggestions.push_back(pattern[pattern_index].fixed_word);
            continue_outer_loop = true;
          } else if (input[arg_index] == pattern[pattern_index].fixed_word) {
            ++arg_index;
            ++pattern_index;
          } else {
            continue_outer_loop = true;
          }
          break;
        case PatternType::AnyWord:
          ++arg_index;
          ++pattern_index;
          break;
        case PatternType::AnyWords:
          ++pattern_index;

          if (pattern_index == pattern.size()) {
            arg_index = input.size();
          } else {
            while (arg_index <= input.size() &&
                   (arg_index == input.size() ||
                    input[arg_index] != pattern[pattern_index].fixed_word)) {
              if (arg_index == input.size()) {
                suggestions.push_back(pattern[pattern_index].fixed_word);
              }
              ++arg_index;
            }
          }
          break;
      }
      if (continue_outer_loop) {
        break;
      }
    }
  }

  // std::sort(suggestions.begin(), suggestions.end());
  // suggestions.erase(std::unique(suggestions.begin(), suggestions.end()), suggestions.end());
  return suggestions;
}

bool Parser::Match(const std::vector<Pattern>& pattern,
                   const std::vector<std::string>& args) const {
  size_t arg_index = 1;
  size_t pattern_index = 0;

  while (arg_index < args.size() && pattern_index < pattern.size()) {
    switch (pattern[pattern_index].type) {
      case PatternType::FixedWord:
        if (args[arg_index] == pattern[pattern_index].fixed_word) {
          ++arg_index;
          ++pattern_index;
        } else {
          return false;
        }

        break;
      case PatternType::AnyWord:
        ++arg_index;
        ++pattern_index;
        break;
      case PatternType::AnyWords:
        ++pattern_index;
        if (pattern_index == pattern.size()) {
          arg_index = args.size();
        } else {
          while (arg_index < args.size() && args[arg_index] != pattern[pattern_index].fixed_word) {
            ++arg_index;
          }
        }
        break;
    }
  }

  if (pattern_index < pattern.size() && pattern[pattern_index].type == PatternType::AnyWords) {
    pattern_index++;
  }

  return arg_index == args.size() && pattern_index == pattern.size();
}

std::vector<Pattern> MakePattern(
    std::initializer_list<std::variant<const char*, PatternType>> elements) {
  std::vector<Pattern> pattern;

  for (const auto& element : elements) {
    if (std::holds_alternative<const char*>(element)) {
      pattern.push_back({PatternType::FixedWord, std::get<const char*>(element)});
    } else {
      pattern.push_back({std::get<PatternType>(element), ""});
    }
  }

  return pattern;
}
